# pylint: disable=missing-docstring, global-statement, invalid-name
#
# Copyright (C) 2017  Jonas Colmsjö, Claes Strannegård
#

#
# A variant of Q-learning where the agents have sensors and motors. Distal
# stimula (Things) are perceived using boolean sensors and thereby create
# proximal stimula that are perceived by the agent. Sensors can also be combined
# using operators such as AND and SEQ operators (a SEQ b = a is true and then b)
# and thereby creating a network (not fully implemented yet).
#
# Actions are modelled based on motors. Combinations of motors
# that are activaed gives the different actions. Motors can be though of as
# parts of a body of an animal or simply motors on a robot/vehicle.
#
# There are no terminal states/nodes. Status values (between 0 and 1) are used
# instead. The agent has reached a terminal state when one status is zero or
# less
#

# Imports
# =======

import math

from collections import defaultdict
from gzutils.gzutils import Logging

from .agents import Agent


# Setup logging
# =============

DEBUG_MODE = True
l = Logging('network_rl', DEBUG_MODE)


# Network model used by the Network Q-learning Agents
# ===================================================

#
# A variant of a MDP where:
# - actions are generated from motors - (m1:bool, m2:bool, ..., mn:bool)
# - states are represented with sensors - (s1:bool, s2:bool, ..., sn:bool)
# - statuses are used instead of terminal states. Any status less than or equal
#   to zero is equivalent to a terminal state - [(name, float)]
# - init is the initial state of the sensors
#

#
# SensorModel
# -----------
#
# Holds a model mapping sensors to states: {(s1,..., sn): 'state'}
#
# It is possible to create sensors from a mdp. One sensor is created for each
# state. This is mainly used in simple examples/testing.
#

def mdp_to_simple_sensor_model(mdp):
    # One sensor for each state in the MDP
    sensor_template = [False] * len(mdp.states)
    i = 0
    model = {}
    for state in sorted(mdp.states):
        sensor = list(sensor_template)
        sensor[i] = True
        i += 1
        model[tuple(sensor)] = state
    return model

# sensors = [('sensor name', Thing to recognise)]
# model   = {(s1:bool,...,sn:bool): 'state'} maps sensor tuples to states (for readability)
class SensorModel:
    # pylint: disable=too-few-public-methods
    def __init__(self, sensors, model=None, mdp=None):
        self.sensors = sensors
        if model and mdp:
            raise Exception('Both model and mdp should not be used!')
        if model:
            self.model = model
        if mdp:
            self.model = mdp_to_simple_sensor_model(mdp)

    def __repr__(self):
        return 'SensorModel:' + str(self.model)

    # return the state for a tuple of sensors
    def __call__(self, sensors):
        if not sensors:
            return None
        return self.model[sensors]

    # returns the sensors tuple for a state
    def sensors_for_state(self, state):
        for k, v in self.model.items():
            if v == state:
                return k
        raise Exception('SensorModel.sensors: state not found - ' + state)


#
# MotorModel
# ----------
#
# motors - list of motors represented with string ['m1',...,'mn']
# model - maps tuples of motors (bool,...,bool) to actions in the Environemtn used:
#        [ ((m1:bool,m2:bool,...,mn:bool), 'action'),...,('*', 'default_actiom') ]
#        The default action is used when a combination of active motors do not
#        result in any action in the environment
#
def find_in_list_with_tuples(key, default_key, idx, lst):
    r = list(filter(lambda x: x[idx] == key, lst))
    if not r:
        r = list(filter(lambda x: x[idx] == default_key, lst))
    return r

class MotorModel:
    # pylint: disable=too-few-public-methods
    def __init__(self, motors, model):
        self.model = model
        self.motors = motors

    # returns the action for a motors tuple
    def __call__(self, motors):
        if not motors:
            return None
        act = find_in_list_with_tuples(motors, '*', 0, self.model)
        if act:
            return act[0][1]
        return None

    def __repr__(self):
        return 'MotorModel:' + str(self.model)


#
# NetworkDP
# ---------
#
# Holds the network used by the NetworkAgents. It consists of motors, sensors
# and statuses. The SensorModel and MotorModel classes are used to manage the
# sensors and motors (data structure classes).
#
# init         = (s1_init, ..., sn_init) - initial state of sensors
# statuses     = {name1: init_value,..., name_n: init_value}
# motor_model  = {(m1:bool,m2:bool,...,mn:bool): 'action name', '*': 'default_action'}
# sensor_model = {(s1:bool, s2:bool, ..., sn:bool): 'state name' } (Optional)
#
class NetworkDP:
    # pylint: disable=too-many-arguments, too-many-instance-attributes
    def __init__(self, init, statuses, motor_model, gamma=.9, sensor_model=None):
        self.init = init
        self.gamma = gamma
        self.statuses = statuses
        self.motor_model = motor_model
        self.sensor_model = sensor_model
        self.init_statuses = dict(statuses)

        self.actlist = self.actions_from_motors()

        self.history = []
        self.history_headers = [str(list(statuses.keys())), 'state', 'action',
                                str(list(statuses.keys()))]

    def __repr__(self):
        return ('gamma=' + str(self.gamma) + '\n' +
                'init=' + str(self.init) +
                ('(' + self.sensor_model(self.init) if self.sensor_model else '') + ')\n' +
                'sensors=' + str(self.sensor_model.sensors) + '\n' +
                'statuses=' + str(self.statuses) + '\n' +
                'motors=' + str(self.motor_model.motors) + '\n' +
                'motor_model=' + str(self.motor_model) + '\n' +
                'sensor_model=' + str(self.sensor_model) + '\n' +
                'actlist=' + str(self.actlist) + '\n' +
                'actions (using motor_model)=' +
                str([self.motor_model(act) for act in self.actlist]) + '\n' +
                'in_terminal=' + str(self.in_terminal()))

    def update_statuses(self, state, action, rewards):
        if self.motor_model:
            action = self.motor_model(action)
        if self.sensor_model:
            state = self.sensor_model(state)
        self.history.append((list(self.statuses.values()), state, action, list(rewards.values())))

        for k, v in rewards.items():
            self.statuses[k] += v

    def reset(self):
        self.statuses = dict(self.init_statuses)

    def in_terminal(self):
        return min([v for k, v in self.statuses.items()]) <= 0

    #
    # motors=[m1,m2,...,mn] actions=[(m1=T/F,m2=T/F,...,mn=T/F)]
    # actions are tuples where each motor is represented by one position with a
    # boolean indicating if it is active.
    #
    # Generate all possible actions from a list of motors (2^n combinations)
    # [m1,m2,...,mn] -> [(F,F,...,F),...,(T,T,...,T])]
    def actions_from_motors(self):
        def ext(e, lst):
            for v in lst:
                v.insert(0, e)

        def actions_(motors):
            if len(motors) == 1:
                return [[True], [False]]
            a = actions_(motors[1:])
            ext(False, a)
            b = actions_(motors[1:])
            ext(True, b)
            a.extend(b)
            return a

        return list(map(tuple, actions_(self.motor_model.motors)))

# Keeps track of the history of actions performed
class NetworkAgent(Agent):
    # pylint: disable=too-many-instance-attributes
    def __init__(self, program=None, name='nonname', ndp=None, max_iterations=None):

        self.program = program or self

        self.ndp = ndp
        self.all_act = ndp.actlist

        # (s)tate, (p)revious (s), (a)ction, (p)revious (a), (r)eward, (p)revious (r)
        self.s = self.ps = None
        self.a = self.pa = None
        self.r = self.pr = {}

        self.iterations = 0
        self.in_terminal = False
        self.max_iterations = max_iterations

        # history of statuses, state, action, reward
        self.history = []
        self.history_headers = [str(list(ndp.statuses.keys())), 'state', 'action',
                                str(list(ndp.statuses.keys()))]

        # history of statuses only
        self.status_history = {}
        for status in ndp.statuses:
            self.status_history[status] = []

        super().__init__(name, self.program)

    def __repr__(self):
        return ('statuses:' + str(self.ndp.statuses) +
                ',iterations:' + str(self.iterations) +
                ',in_terminal:' + str(self.in_terminal))

    def reset(self):
        self.iterations = 0
        self.in_terminal = False
        self.ndp.reset()

    # keep count of the number of iterations and check if the limit is reached
    def check_iterations(self):
        self.iterations += 1
        return self.max_iterations and self.iterations >= self.max_iterations

    # Check if status is zero or less and keep track of number of
    # iterations and stop after some limit has been reached
    # The iterations counter is increased so the function should be called once
    def check_terminal(self):
        self.alive = not (self.ndp.in_terminal() or self.check_iterations())
        return not self.alive

    def actions_in_state(self, state):
        # pylint: disable=unused-argument
        return self.all_act

    # update the status after the action has been perfomred and the new state reached
    def update_statuses(self):
        s, a, r = self.s, self.a, self.r
        motor_model, sensor_model, statuses = (self.ndp.motor_model, self.ndp.sensor_model,
                                               self.ndp.statuses)

        # save history for statuses, state, action, reward
        if motor_model:
            a = motor_model(a)
        if sensor_model:
            s = sensor_model(s)
        self.history.append((list(statuses.values()), s, a, list(r.values())))

        # save status history
        for status in self.ndp.statuses:
            self.status_history[status].append(r[status])

        self.ndp.update_statuses(s, a, r)
        self.ps, self.pa, self.pr = s, a, dict(r)

    # to be used after the __call__ function
    def any_status_increased(self):
        i, res = self.iterations, False
        if i < 2:
            return False
        for status in self.ndp.statuses:
            res = res or self.status_history[status][i] > self.status_history[status][i-1]
        return res


    # To be overridden when not supplying the program in the constructor.
    # Environment expects a program function with the type:
    # action = program(percept=([string|Thing|NSArtifact], reward for previous actopn))
    def __call__(self, percept):
        a1 = None
        s1, r = self.update_state(percept)
        if self.program:
            a1 = self.program(percept)

        self.s, self.a, self.r = s1, a1, r
        self.update_statuses()
        return a1

    # To be overridden in most cases. The default case assumes the percept
    # to be of type (state, reward)
    def update_state(self, percept):
        # pylint: disable=no-self-use
        return percept


#
# NetworkQLearningAgent
# =====================
#
# Implements multi-dimensional Q-learning using the NetworkDP class instead of a MDP.
#
class NetworkQLearningAgent(NetworkAgent):
    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(self, ndp, Ne, Rplus, alpha=None, delta=0.5, max_iterations=None, name='noname'):

        # Multidimensional Q: Q_status[s, a]
        self.Q = {}
        for status in ndp.statuses:
            self.Q[status] = defaultdict(float)

        self.Ne = Ne                      # iteration limit in exploration function
        self.Nsa = defaultdict(float)
        self.delta = delta
        self.Rplus = Rplus                # large value to assign before iteration limit
        self.gamma = ndp.gamma

        if alpha:
            self.alpha = alpha
        else:
            self.alpha = lambda n: 1./(1+n)  # udacity video

        super().__init__(None, name, ndp, max_iterations)

    def __repr__(self):
        res = ''
        for status in self.ndp.statuses:
            lst = sorted([(self.ndp.sensor_model(k[0]) if k[0] else 'None',
                           self.ndp.motor_model(k[1]), v) for k, v in self.Q[status].items()],
                         key=lambda x: x and x[0])
            lst = list(filter(lambda x: x[2] != 0.0, lst))
            res += status + ':' + str(lst)
        return ('Q:' + res  +
                ',statuses:' + str(self.ndp.statuses) +
                ',iterations:' + str(self.iterations) +
                ',in_terminal:' + str(self.in_terminal))

    def Q_to_U_and_pi(self):
        res = {}
        for status in self.ndp.statuses:
            U = defaultdict(lambda: -math.inf) # Very Large Negative Value for Comparison see below.
            pi = {}
            for state_action, value in self.Q[status].items():
                state, action = state_action
                if U[state] < value:
                    U[state] = value
                    pi[state] = action
            U = dict(map(lambda x: (self.ndp.sensor_model(x[0]), x[1]), U.items()))
            pi = dict(map(lambda x: (self.ndp.sensor_model(x[0]),
                                     self.ndp.motor_model(x[1])), pi.items()))
            res[status] = U, pi
        return res

    # check if status is zero or less and keep track of number of
    # iterations and stop after some limit has been reached
    def check_terminal(self):
        self.in_terminal = self.ndp.in_terminal() or self.check_iterations()
        return self.in_terminal

    # Exploration function. Returns fixed Rplus untill agent has visited state,
    # action a Ne number of times.
    def f(self, u, n):
        if n < self.Ne:
            return self.Rplus
        return u

    def visited_states(self):
        return sorted(set(map(lambda x: x[0], list(self.Q))))

    def __call__(self, percept):
        # pylint: disable=too-many-locals
        s1, r = self.update_state(percept)
        Q, Nsa, s, a = self.Q, self.Nsa, self.s, self.a
        alpha, gamma, delta, in_terminal = self.alpha, self.gamma, self.delta, self.check_terminal()
        actions_in_state, statuses = self.actions_in_state, self.ndp.statuses

        if in_terminal:
            for objective in statuses:
                Q[objective][(s, None)] = r[objective]
        if s is not None:
            Nsa[s, a] += 1
            for objective in statuses:
                Q[objective][(s, a)] += (alpha(Nsa[s, a]) *
                                         (r[objective] +
                                          gamma * max([Q[objective][(s1, a1)]
                                                       for a1 in actions_in_state(s1)]) -
                                          Q[objective][(s, a)]))

        if in_terminal:
            self.s = self.a = self.r = None
        else:
            self.s, self.r = s1, r

            best_action = {}
            for objective in statuses:
                best_action[objective] = max([(self.f(statuses[objective] +
                                                      delta * Q[objective][(s1, a1)],
                                                      Nsa[s1, a1]), a1)
                                              for a1 in actions_in_state(s1)],
                                             key=lambda x: x[0])
            self.a = min(list(best_action.items()), key=lambda x: x[1][0])[1][1]

            # original version
            # self.a = argmax(actions_in_state(s1),
            #                 key=lambda a1: self.f(Q[objective][(s1, a1)], Nsa[s1, a1]))

            self.ndp.update_statuses(self.s, self.a, self.r)
        return self.a


#
# LocalQLearningAgent
# ===================
#
# An agent where Q-values are connected to the nodes in the network rather than
# to states. The boolean nodes in the network can be seen as small MDPs. The
# transition model (with correspodning probabilities and rewards) might be unknow,
# but this does not differ compared to ordinary Q-learning. If the environment
# the agent lives in can be modelled with an MDP, then the sensors will also model
# this MDP (perhaps not completely depending on if the sensors covers all aspects of
# the environment).
#
class LocalQLearningAgent:
    pass



# pylint: disable=missing-docstring, global-statement, invalid-name, too-few-public-methods
#
# Copyright (C) 2017  Jonas Colmsjö, Claes Strannegård
#


# Imports
# ======

import unittest
from gzutils.gzutils import Logging
from ecosystem.network import Network, MotorNetwork


# Setup logging
# =============

DEBUG_MODE = True
l = Logging('test_network', DEBUG_MODE)


class Thing1:
    pass

class Thing2:
    pass

class Thing3:
    pass

class TestNetwork(unittest.TestCase):

    def setUp(self):
        l.info('Testing network...')

    def tearDown(self):
        l.info('...done with test_network.')

    def test_SENSOR_AND(self):
        network = Network()
        network.add_SENSOR_node(Thing1)

        self.assertTrue(network.get_state() == (None,))

        network.update([(Thing2(), 1.0)])
        self.assertTrue(network.get_state() == (False,))

        network.update([(Thing1(), 1.0)])
        self.assertTrue(network.get_state() == (True,))

        network.add_SENSOR_node(Thing2)
        network.update([(Thing2(), 1.0)])
        self.assertTrue(network.get_state() == (False, True, ))

        network.update([])
        self.assertTrue(network.get_state() == (False, False, ))

        network.update([(Thing2(), 2.0), (Thing1(), 0.5)])
        self.assertTrue(network.get_state() == (True, True, ))

        network.add_AND_node([0, 1])
        network.update([(Thing2(), 1.0), (Thing1(), 0.3)])
        self.assertTrue(network.get_state() == (True, True, True))

        network.update([(Thing2(), 2.0)])
        self.assertTrue(network.get_state() == (False, True, False))

    def test_SEQ(self):
        network = Network([('thing1', Thing1), ('thing2', Thing2)])
        network.add_SEQ_node([0, 1])

        network.update([(Thing1(), 1.0)])
        self.assertTrue(network.get_state() == (True, False, False))

        network.update([(Thing2(), 0.5)])
        self.assertTrue(network.get_state() == (False, True, True))

        network.update([(Thing2(), 1.0)])
        self.assertTrue(network.get_state() == (False, True, False))

        network.update([])
        self.assertTrue(network.get_state() == (False, False, False))

    def test_top_active(self):
        network = Network()
        n1 = network.add_SENSOR_node(Thing1)
        n2 = network.add_SENSOR_node(Thing2)
        n3 = network.add_SENSOR_node(Thing3)
        n4 = network.add_SEQ_node([n1, n2, n3])

        network.update([(Thing1(), 1.0)])
        self.assertTrue(network.get() == set([n1]))
        self.assertTrue(network.top_active() == set([n1]))

        network.update([(Thing2(), 1.0)])
        self.assertTrue(network.get() == set([n2]))
        self.assertTrue(network.top_active() == set([n2]))

        network.update([(Thing3(), 1.0)])
        self.assertTrue(network.get() == set([n3, n4]))
        self.assertTrue(network.top_active() == set([n4]))

        network.update([])
        self.assertTrue(network.get() == set())

        # Check that SEQ can keep track of multiple sequences at the same time
        network.update([(Thing1(), 1.0)])
        network.update([(Thing2(), 1.0), (Thing1(), 1.0)])
        network.update([(Thing3(), 1.0), (Thing2(), 1.0)])
        self.assertTrue(network.get() == set([n3, n2, n4]))
        network.update([(Thing3(), 1.0)])
        self.assertTrue(network.get() == set([n3, n4]))

    def test_RAND_NOT(self):
        network = Network()
        n1 = network.add_RAND_node(0.5)
        n2 = network.add_NOT_node([n1])
        for _ in range(0, 100):
            network.update([])
            state = network.get_state()
            self.assertTrue(state[n1] != state[n2])

    def test_ONE(self):
        network = Network()
        n1 = network.add_SENSOR_node(Thing1)
        n2 = network.add_SENSOR_node(Thing2)
        n3 = network.add_SENSOR_node(Thing3)
        n1T = network.add_ONE_node(n1, [n1, n2, n3])
        n2T = network.add_ONE_node(n2, [n1, n2, n3])
        n3T = network.add_ONE_node(n3, [n1, n2, n3])

        self.assertTrue(network.update([(Thing1(), 1.0)]) - set([n1]) == set([n1T]))
        self.assertTrue(network.update([(Thing2(), 1.0)]) - set([n2]) == set([n2T]))
        self.assertTrue(network.update([(Thing3(), 1.0)]) - set([n3]) == set([n3T]))
        self.assertTrue(network.update([(Thing1(), 1.0), (Thing2(), 1.0)]) -
                        set([n1, n2]) == set())
        self.assertTrue(network.update([(Thing1(), 1.0), (Thing2(), 1.0), (Thing3(), 1.0)]) -
                        set([n1, n2, n3]) == set())
        self.assertTrue(network.update([]) == set())

        m1 = network.add_MIN_node(1, [n1, n2, n3])
        m2 = network.add_MIN_node(2, [n1, n2, n3])
        m3 = network.add_MIN_node(3, [n1, n2, n3])

        vs = network.update([(Thing3(), 1.0)])
        self.assertTrue(m1 in vs and not m2 in vs and not m3 in vs)
        vs = network.update([(Thing1(), 1.0), (Thing2(), 1.0)])
        self.assertTrue(m2 in vs and m1 in vs and not m3 in vs)
        vs = network.update([(Thing1(), 1.0), (Thing2(), 1.0), (Thing3(), 1.0)])
        self.assertTrue(m3 in vs and  m2 in vs and  m1 in vs)

        m1 = network.add_MAX_node(1, [n1, n2, n3])
        m2 = network.add_MAX_node(2, [n1, n2, n3])
        m3 = network.add_MAX_node(3, [n1, n2, n3])

        vs = network.update([])
        self.assertTrue(m1 in vs and m2 in vs and  m3 in vs)
        vs = network.update([(Thing3(), 1.0)])
        self.assertTrue(m1 in vs and m2 in vs and  m3 in vs)
        vs = network.update([(Thing1(), 1.0), (Thing2(), 1.0)])
        self.assertTrue(m2 in vs and not m1 in vs and m3 in vs)
        vs = network.update([(Thing1(), 1.0), (Thing2(), 1.0), (Thing3(), 1.0)])
        self.assertTrue(m3 in vs and not m2 in vs and not m1 in vs)

        l.debug('network.__repr__:', network)
        l.debug('network.toGraphviz:', network.toGraphviz())
        network.saveGraphviz('graph.dot')


class TestMotorNetwork(unittest.TestCase):

    def setUp(self):
        l.info('Testing MotorNetwork...')

    def tearDown(self):
        l.info('...done with MotorNetwork.')

    def test_MOTOR_AND_SEQ(self):
        mnetwork = MotorNetwork()
        n1 = mnetwork.add_MOTOR_node('m1')
        self.assertTrue(mnetwork.get() == set())

        mnetwork.update(set([n1]))
        self.assertTrue(mnetwork.get() == set([0]))

        mnetwork.update(set())
        self.assertTrue(mnetwork.get() == set())

        n2 = mnetwork.add_MOTOR_node('m2')
        mnetwork.update(set([n2]))
        self.assertTrue(mnetwork.get() == set([n2]))

        mnetwork.update(set([n2, n1]))
        self.assertTrue(mnetwork.get() == set([n1, n2]))

        n3 = mnetwork.add_MAND_node([n1, n2])
        mnetwork.update(set())
        self.assertTrue(mnetwork.get() == set())
        mnetwork.update(set([n3]))
        self.assertTrue(mnetwork.get() == set([n1, n2]))

        n4 = mnetwork.add_MOTOR_node('m3')
        n5 = mnetwork.add_MSEQ_node([n3, n4])
        mnetwork.update(set())
        self.assertTrue(mnetwork.get() == set())
        mnetwork.update(set())
        self.assertTrue(mnetwork.get() == set())
        mnetwork.update(set([n5]))
        self.assertTrue(mnetwork.get() == set([n1, n2]))
        mnetwork.update(set())
        self.assertTrue(mnetwork.get() == set([n4]))
        mnetwork.update(set())
        self.assertTrue(mnetwork.get() == set())

        mnetwork.delete_nodes([n5])
        mnetwork.update(set([n4]))
        self.assertTrue(mnetwork.get() == set([n4]))

    def test_motor_to_action(self):

        # set([(active_motors, action)])
        # each motor is mapped to one action. It is only meaningful to
        # activate one motor at the time. All other combinations of
        # active motors are mapped to the action '-' (do nothing)
        motors = ['<', '>', '^', 'v']
        motors_to_action1 = {frozenset([0]): '<',
                             frozenset([1]): '>',
                             frozenset([2]): '^',
                             frozenset([3]): 'v',
                             '*': '-'}

        mnetwork = MotorNetwork(motors, motors_to_action1)
        mnetwork.update(frozenset([0]))
        self.assertTrue(mnetwork.get_action() == '<')
        mnetwork.update(frozenset([1]))
        self.assertTrue(mnetwork.get_action() == '>')
        mnetwork.update(frozenset([2]))
        self.assertTrue(mnetwork.get_action() == '^')
        mnetwork.update(frozenset([3]))
        self.assertTrue(mnetwork.get_action() == 'v')


        # Motor model where all combinations of active motors have an action
        motors_to_action2 = {frozenset(): '<',
                             frozenset([0]): '>',
                             frozenset([1]): '^',
                             frozenset([0, 1]): 'v'}

        mnetwork = MotorNetwork(motors, motors_to_action2)
        mnetwork.update(frozenset())
        self.assertTrue(mnetwork.get_action() == '<')
        mnetwork.update(frozenset([0]))
        self.assertTrue(mnetwork.get_action() == '>')
        mnetwork.update(frozenset([1]))
        self.assertTrue(mnetwork.get_action() == '^')
        mnetwork.update(frozenset([0, 1]))
        self.assertTrue(mnetwork.get_action() == 'v')



# pylint: disable=missing-docstring, global-statement, invalid-name
#

# Imports
# =======

import math
import os
from random import random
from gzutils.gzutils import Logging, get_output_dir


# Setup logging
# =============

DEBUG_MODE = True
l = Logging('network', DEBUG_MODE)

#
# type - a string speciying the type of node (for information only)
# func - the function to apply to the percept:[(Thing|NonSpatial, radius)],
#        vars:[any type] -> (boolean, [any type])
# vars - the initial state the function executes in, necessary for SEQ (and other
#        function that require a state)
# children - indexes if the nodes that the nodes takes input from (used in the update function)
#
class Node:
    # pylint: disable=too-few-public-methods
    def __init__(self, type_, func, vars_, children):
        self.type_ = type_
        self.vars_ = vars_
        self.func = func
        self.children = children

    def __call__(self, percept):
        res, self.vars_ = self.func(percept, self.vars_)
        return res

    def __repr__(self):
        return ('(' + self.type_ + ' - vars:' + str(self.vars_) +
                ',children:' + str(self.children) + ')')

# Create a sensor that recognise Things of type `cls`, radius is ignored
def SENSOR_factory(cls):
    return Node('SENSOR:' + cls.__name__,
                lambda things, _: (any([isinstance(x, cls) for x, _ in things]), []),
                [], [])

def AND_factory(indexes, state):
    return Node('AND', lambda _, _2: (all([state[i] for i in indexes]), []),
                [], indexes)

def OR_factory(indexes, state):
    return Node('OR', lambda _, _2: (any([state[i] for i in indexes]), []),
                [], indexes)

def ONE_factory(index, indexes, state):
    return Node('ONE', lambda _, _2: (all([state[index] if i == index else not state[i] for i in indexes]), []),
                [], indexes)

def MIN_factory(n, indexes, state):
    return Node('MIN', lambda _, _2: ([state[i] for i in indexes].count(True) >= n, []),
                [], indexes)

def MAX_factory(n, indexes, state):
    return Node('MAX', lambda _, _2: ([state[i] for i in indexes].count(True) <= n, []),
                [], indexes)

# _=indexes, _2=state
def RAND_factory(prob):
    return Node('RAND', lambda _, _2: (random() < prob, []),
                [], [])

def NOT_factory(indexes, state):
    return Node('NOT', lambda _, _2: (all([not state[i] for i in indexes]), []),
                [], indexes)


# t:state[idx1] followed by t+1:state[idx2]
# NOTE: the value of t:state[idx2] and t+1:state[idx1] makes no difference
#def SEQ_factory(idx1, idx2, state):
#    return Node('SEQ',
#                lambda things, vars_: (bool(vars_[0]) and state[idx2], [state[idx1]]),
#                [None], [idx1, idx2])

# Create a counter initially set to 0 each time state[index] is true.
# Update state[children[counter]] each time update is called.
# Increase the counter until all children have been updated, then delete it.
# Many counters can exist at the same time.
# vars = [counter1, ..., countern]
def SEQ_factory(children, state):
    def update(_, vars_):
        # the (seq, counter) tuples are stored in reverse order in vars_
        if state[children[0]]:
            vars_.insert(0, (True, 0))
        new_vars = []
        seq = False
        counter = math.inf
        for seq, counter in vars_:
            seq = seq and state[children[counter]]
            if counter < len(children) - 1:
                new_vars.append((seq, counter + 1))

        return (seq and counter == len(children) - 1, new_vars)
    return Node('SEQ', update, [], children)


#
# state - a list of booleans consisting of the sensors and other nodes that
#         have been added to the network (AND, SEQ etc.)
# nodes - a list of Node objects used to calculate the values for the SENSORS and
#         other types of nodes (AND, SEQ etc.)
# root_nodes - the root Nodes in the trees that makes up the network
#
class Network:

    # sensors = [('sensor name', Thing to recognise)]
    def __init__(self, sensors=None):
        self.state = []
        self.nodes = []
        self.root_nodes = []
        if sensors:
            self.add_sensors(sensors)

    def add_sensors(self, sensors):
        for _, cls in sensors:
            self.add_SENSOR_node(cls)

    def __repr__(self):
        res = ''
        for i in range(0, len(self.nodes)):
            node = self.nodes[i]
            res += node.type_ + ':' + str(self.state[i]) + ':' + str(node.children) + ';'
        return res

        return (res +
                'state:' + str(self.state) +
                ', nodes:' + str(self.nodes) +
                ', root_nodes:' + str(self.root_nodes))


    def toGraphviz(self, color_edges=False):
        colors = ['black', 'blue', 'brown', 'cyan', 'darkgreen', 'deeppink', 'gold']
        res = 'digraph G {\n\tsize ="8,8";\n'
        for i in range(0, len(self.nodes)):
            res += '\t{} [label="{}\\n{}\\n{}"];\n'.format(i, i,
                                                           self.nodes[i].type_,
                                                           self.state[i])
            res += '\tedge [color={}];\n'.format(colors[i % len(colors)]) if color_edges else ''
            for child in self.nodes[i].children:
                res += '\t{}->{};\n'.format(child, i)
        res += '}'
        return res

    def saveGraphviz(self, filename, output_dir=None):
        # Save the performance history to file
        if not output_dir:
            output_dir = get_output_dir()
        output_path = os.path.join(output_dir, filename)
        l.debug(output_path)
        filep = open(output_path, 'w')
        print(self.toGraphviz(), file=filep)
        filep.close()

    # the state of the network is updated using a depth first search starting
    # in the root nodes. Nodes might be updated several times, but this will do
    # for now.
    def update(self, percept):
        for node in self.root_nodes:
            self.update_node(node, percept)
        return self.get()

    def update_node(self, node, percept):
        idx = self.nodes.index(node)
        for i in node.children:
            self.update_node(self.nodes[i], percept)
        self.state[idx] = node(percept)

    def get_state(self):
        return tuple(self.state)

    # return a set of indexes for the nodes that are active
    def get(self):
        res = set()
        for i in range(0, len(self.state)):
            if self.state[i]:
                res |= {i}
        return frozenset(list(res))

    def add_root_node(self, node):
        self.state.append(None)
        self.nodes.append(node)
        self.root_nodes.append(node)
        return len(self.state) - 1

    # when other nodes than SENSORs are added are the nodes below no longer
    # root nodes
    def delete_root_nodes(self, indexes):
        for i in indexes:
            if self.nodes[i] in self.root_nodes:
                self.root_nodes.remove(self.nodes[i])

    def add_SENSOR_node(self, cls):
        return self.add_root_node(SENSOR_factory(cls))

    def add_RAND_node(self, prob):
        return self.add_root_node(RAND_factory(prob))

    def add_AND_node(self, indexes):
        idx = self.add_root_node(AND_factory(indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

    def add_OR_node(self, indexes):
        idx = self.add_root_node(OR_factory(indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

    def add_ONE_node(self, index, indexes):
        idx = self.add_root_node(ONE_factory(index, indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

    def add_MIN_node(self, n, indexes):
        idx = self.add_root_node(MIN_factory(n, indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

    def add_MAX_node(self, n, indexes):
        idx = self.add_root_node(MAX_factory(n, indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

    def add_NOT_node(self, indexes):
        idx = self.add_root_node(NOT_factory(indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

#    def add_SEQ_node(self, idx1, idx2):
#        idx = self.add_root_node(SEQ_factory(idx1, idx2, self.state))
#        self.delete_root_nodes([idx1, idx2])
#        return idx

    def add_SEQ_node(self, indexes):
        idx = self.add_root_node(SEQ_factory(indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

    def delete_nodes(self, indexes):
        self.delete_root_nodes(indexes)
        for idx in indexes:
            for child in self.nodes[idx].children:
                self.root_nodes.append(self.nodes[child])
            self.nodes[idx] = self.state[idx] = None

    def top_active_(self, node):
        idx = self.nodes.index(node)
        if self.state[idx]:
            return {idx}
        res = set()
        for child in node.children:
            res |= self.top_active_(self.nodes[child])
        return res

    def top_active(self, percept=None):
        if percept:
            self.update(percept)
        res = set()
        for node in self.root_nodes:
            res |= self.top_active_(node)
        return frozenset(list(res))


#
# MotorNetwork
#
# Similar to Network but with MOTORS in the leafes. The network nodes can be
# MAND, MSEQ etc. Nodes are added in the same way as in the Network class
# (and the root nodes updated as more nodes are added). Updating starts with
# a tuple representing the state for the root nodes.#
#
# TODO: Should separate out the common parts rather than inheriting Network
#       (add_SENSOR_node etc. makes little sense in this class)
#

# The Node class is used in a different way here.
#
# func - a function that updates the state of the nodes. The only input is the
#        current state of the nodes which has been set by the update method.
#        (vars: [any type] -> [])

# Create a motor
def MOTOR_factory(name):
    return Node('MOTOR:' + name,
                lambda _, _2: (None, []),
                [], [])

def MAND_factory(index, children, state):
    def update(_, _2):
        for idx in children:
            state[idx] = state[index]
        return (None, [])
    return Node('MAND', update, [], children)

# Create a counter initially set to 0 each time state[index] is true.
# Update state[children[counter]] each time update is called.
# Increase the counter until all children have been updated, then delete it.
# Many counters can exist at the same time.
# vars = [counter1, ..., countern]
def MSEQ_factory(index, children, state):
    def update(_, vars_):
        if state[index]:
            vars_.append(0)
        new_vars = []
        for counter in vars_:
            state[children[counter]] = True
            if counter < len(children) - 1:
                new_vars.append(counter + 1)

        return (None, new_vars)
    return Node('MSEQ', update, [], children)


class MotorNetwork(Network):

    # motors = ['motor name']
    def __init__(self, motor_names=None, motors_to_action=None):
        super().__init__()

        motor_names = motor_names or []

        # indexes of each MOTOR in self.state
        self.motors = []
        self.motor_names = []
        self.motors_to_action = motors_to_action or {'*': '-'}

        for name in motor_names:
            self.add_MOTOR_node(name)

    def __repr__(self):
        return ('state:' + str(self.state) +
                ', motor_names:' + str(self.motor_names) +
                ', nodes:' + str(self.nodes) +
                ', root_nodes:' + str(self.root_nodes) +
                ', motors:' + str(self.motors))

    # indexes is a set with the indexes of the nodes that should be True
    def update(self, percept):
        # pylint disable=arguments-differ
        for idx in range(0, len(self.state)):
            self.state[idx] = idx in percept
        for node in self.root_nodes:
            self.update_node(node, None)
        return self.get_action()

    def update_node(self, node, percept):
        node(None)
        for i in node.children:
            self.update_node(self.nodes[i], None)

    # return a (frozen)set of indexes for the motors that are active
    def get(self):
        res = set()
        for i in self.motors:
            if self.state[i]:
                res |= {i}
        return frozenset(list(res))

    def get_action(self):
        action = self.motors_to_action.get(self.get(), None)
        return action or self.motors_to_action['*']

    def add_MOTOR_node(self, name):
        self.motor_names.append(name)
        idx = self.add_root_node(MOTOR_factory(name))
        self.motors.append(idx)
        return idx

    def add_MAND_node(self, indexes):
        idx = self.add_root_node(MAND_factory(len(self.state), indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx

    def add_MSEQ_node(self, indexes):
        idx = self.add_root_node(MSEQ_factory(len(self.state), indexes, self.state))
        self.delete_root_nodes(indexes)
        return idx
